# -*- coding: utf-8 -*-
"""
Add some properties to DBObject, e.g. full instantiation through __init__.
"""

from collections import OrderedDict
import itertools
import os
from typing import List

from astropy.io import fits
from astropy.io.fits import ImageHDU

from common.database.ClassCache import classcache
from common.database.DataObject import DataObject, persistent
from common.database.DBSelect import Select
from common.database.DBMeta import DBObjectMeta as DBObjectMeta
from common.database.DBProperties import link_property

# TODO Fix this dirty hack with the KEYWORDS_TEMPLATE
# These are keywords that are *ONLY* part of templates.
KEYWORDS_TEMPLATE = {
    "DET DIT",
    "DET NDIT",
    "DET READOUTMODE",
    "INS MODE",
    "DPR CATG",
    "DPR TECH",
    "DPR TYPE",
}


class FakePropertyDict(dict):
    """
    Based on oraclesupport
    TODO: Implement in DBFilebased.
    """
    def __init__(self, aclass):
        """init"""
        self.aclass = aclass


class MDBObjectMeta(DBObjectMeta):
    """
    Enhanced DBObjectMeta for MICADO.
    """

    def select_all(cls):
        """
        Should be implemented by instantiations of MDBObjectMeta

        Returns
        -------
        Select
            Selecting all objects of this class.
        """
        raise NotImplemented

    def register(cls):
        """
        Overloaded to always fill classcache.
        """
        super().register()
        if cls not in classcache:
            classcache[cls] = FakePropertyDict(cls)

    def __call__(cls, autocommit=True, *args, **kwargs):
        """
        Initialize a DBObject.

        Parameters
        ----------
        autocommit: Bool
            The object will automatically be committed if props are given.
        """
        # TODO: Refactor this monstrosity in smaller functions.
        args_fits = ['data', 'header']

        if 'CLS' in kwargs:
            # The CLS is the name of a derived class. See Template
            cls_id = kwargs['CLS']
            kwargs2 = {k: v for k, v in kwargs.items() if k not in ['CLS']}
            result = classcache.get(cls_id, None)
            if result is None:
                print(f"Warning: cannot find class {cls_id}")
                cls_new = cls
            else:
                cls_new = result.aclass
            s = cls_new.__call__(autocommit=autocommit, *args, **kwargs2)
        elif 'object_id' in kwargs:
            s = DBObjectMeta.__call__(cls, *args, **kwargs)
        elif kwargs.keys() == {'filename'}:
            filename = kwargs['filename']
            # A hash of the filename is linked to the filename
            # so the filename can be used in fits headers.
            if os.path.islink(filename) and not filename.endswith(".fits"):
                filename = os.readlink(filename)

            # Assume it's a fits file
            hdulist = fits.open(filename)
            hduprimary = hdulist[0]

            hdus: List[fits.ImageHDU] = [hdu for hdu in hdulist[1:] if isinstance(hdu, fits.ImageHDU)]
            if hasattr(cls, 'extensions'):
                prop = getattr(cls, 'extensions')
                extensions = [
                    prop.prop_type(header=hdu.header)
                    for hdu in hdus
                ]
                hdu: ImageHDU
                for extension, hdu in zip(extensions, hdus):
                    extension.image = hdu.data

                s = cls(autocommit=autocommit, header=hduprimary.header, extensions=extensions, filename=filename)
            else:
                s = cls(autocommit=autocommit, header=hduprimary.header, filename=filename)

            # TODO: Fix filename is ignored stuff
            s.filename = filename
        elif 'header' in kwargs and 'CLS' in kwargs['header']:
            # TODO: ESO uses ID where we use CLS, switch to that.
            # Need to switch classes
            header = kwargs['header']
            cls_id = header['CLS']
            header2 = {k: v for k, v in header.items() if k not in ['CLS']}
            kwargs2 = {k: v for k, v in kwargs.items() if k not in ['header']}
            kwargs2['header'] = header2
            kwargs2['CLS'] = cls_id
            s = cls(autocommit=autocommit, *args, **kwargs2)
        elif 'header' in kwargs:
            def fix_key(key):
                if key.startswith("ESO "):
                    key = key[4:]
                # TODO: Fix KEYWORDS_TEMPLATE hack
                if key in KEYWORDS_TEMPLATE:
                    key = f"TPL {key.replace(' ', '__')}"
                return key

            header0 = {
                fix_key(k): v
                for k, v in kwargs['header'].items()
            }
            # The headers might be unsorted, so need to sort to get the groupby to work.
            # Some ugly stuff with templates is done here, so need to do them first.
            keys = sorted(sorted(header0), key=lambda k: ~k.startswith("TPL "))
            header = OrderedDict([(k, header0[k]) for k in keys])

            # TODO: Fix TEMPLATEOBHACK
            template = None

            ig = itertools.groupby(header.items(), key=lambda ki: ki[0].split(maxsplit=1)[0])
            lig = [(k, list(v)) for k, v in ig]
            # kwargs might contain 'extensions', see branch above and also Raw.__init__
            kwargs2 = {k: v for k, v in kwargs.items() if k not in ['header']}
            for k, lv in lig:
                if " " not in lv[0][0]:
                    # Normal item
                    assert len(lv) == 1, "Value to long {lv}"
                    attrs_existing = cls.get_persistent_properties()
                    attrs_existing_max8 = {a[:8]: a for a in attrs_existing}
                    v2 = lv[0][1]
                    if hasattr(cls, k):
                        # e.g. BITPIX
                        k2 = k
                    elif hasattr(cls, k.lower()):
                        # e.g. filename
                        k2 = k.lower()
                    elif k.lower()[:8] in attrs_existing_max8:
                        # e.g. instrument
                        k2 = attrs_existing_max8[k.lower()[:8]]
                    else:
                        print(f"Warning: {cls.__name__} has no {k}")
                        continue

                    # Convert empty cards to None.
                    if isinstance(v2, fits.card.Undefined):
                        v2 = None

                    # TODO fix conversion of '' to None
                    prop = getattr(cls, k2)
                    if v2 == '':
                        if issubclass(prop.prop_type, MObject):
                            v2 = None
                    kwargs2[k2] = v2
                else:
                    # HIERARCH keyword
                    v2 = OrderedDict([
                        (k2i.split(maxsplit=1)[1], v2i)
                        for k2i, v2i in lv
                    ])
                    dd = {
                        a[:3].upper(): a for a in cls.get_persistent_properties()
                    }
                    # TODO: Create better solution for TPL/TEM hack
                    if 'TEM' in dd and 'TPL' not in dd:
                        dd['TPL'] = dd['TEM']
                    ks = k.strip("0123456789")
                    if k in dd:
                        # An object reference.
                        k2 = dd[k]
                        prop = getattr(cls, k2)
                        # TODO: Fix TEMPLATEOBHACK
                        if k == "OBS":
                            assert template is not None
                            inst2 = prop.prop_type(header=v2, templates=[template])
                        else:
                            inst2 = prop.prop_type(header=v2)
                        # TODO: Fix TEMPLATEOBHACK
                        if k == "TPL":
                            template = inst2
                        kwargs2[k2] = inst2
                    elif ks in dd:
                        # A list object reference.
                        k2 = dd[ks]
                        prop = getattr(cls, k2)
                        inst2 = prop.prop_type(header=v2)
                        if k2 not in kwargs2:
                            kwargs2[k2] = []
                        kwargs2[k2].append(inst2)
                    else:
                        print(f"Warning: {cls.__name__} has no {k} {dd}")
            s = cls(autocommit=autocommit, *args, **kwargs2)
        elif kwargs:
            args_do = ['object_id', 'filename', 'globalname']
            attrs_persistent = [
                a for a in cls.get_persistent_properties()
                # TODO: Fix filename stuff, do not inherit from DataObject?
                if not a.startswith('dummy') and a not in args_do
            ]
            props_persistent = {a: getattr(cls, a) for a in attrs_persistent}
            props_with_default = {
                a: p for a, p in props_persistent.items()
                if p.default is not None}

            # For now be very explicit in the necessary properties.
            # The default is used for attributes that are not specified.
            # Perhaps this can be relaxed. E.g. query for objects with any value
            # for unspecified properties.
            params_unknown = [a for a in kwargs if a not in attrs_persistent and a not in args_fits and a not in args_do]
            params_missing = [
                a for a in attrs_persistent
                if a not in kwargs and a not in props_with_default
            ]
            params_default = [a for a in props_with_default if a not in kwargs]
            if params_unknown or params_missing:
                raise AttributeError(
                    "Got unknown incorrect attributes: {} {}".format(
                        params_unknown, params_missing))

            q = cls.select_all()
            for name, value in kwargs.items():
                q &= getattr(cls, name) == value

            lenq = len(q)
            assert lenq in [0, 1], "There should never be more than 1: {}".format(lenq)

            if lenq:
                s = q[0]
            else:
                s1 = cls()
                for name, value in kwargs.items():
                    # Mainly prevents setting None values.
                    if getattr(s1, name) is not value:
                        setattr(s1, name, value)
                for name in params_default:
                    setattr(s1, name, props_with_default[name].default)
                # Allow class to override defaults above
                s1.__init__(*args, **kwargs)
                if autocommit:
                    s1.commit()
                    # Re-retrieve the object to ensure the id is always
                    # identical.
                    if s1.persists():
                        s = cls(object_id=s1.object_id.binary())
                    else:
                        # There probably is no database backend.
                        s = s1
                else:
                    s = s1
        else:
            s = DBObjectMeta.__call__(cls, *args, **kwargs)

        return s

    def get_raw_class(cls):
        """
        Get the main raw class for a processed data item.

        There is no proper base class to put this in.
        """
        from micado.main.Raw import Raw
        to_search = [getattr(cls, an).prop_type for an in cls.get_persistent_properties()]
        to_search.sort(key=lambda a: -100 if issubclass(a, Raw) else 0)
        while to_search:
            attrtype = to_search.pop(0)
            if isinstance(attrtype, tuple):
                # probably (int, long)
                continue
            if issubclass(attrtype, Raw):
                return attrtype
            if issubclass(attrtype, MObject):
                to_search_more = [getattr(attrtype, an).prop_type for an in attrtype.get_persistent_properties()]
                to_search += to_search_more
        return None


class MObject(DataObject, metaclass=MDBObjectMeta):
    """
    Enhanced DBObject.
    """

    dummy9 = persistent("dummy9", str, "")

    @property
    def fits_keywords(self):
        """
        Create FITS keywords.
        """
        kws = self.real_fits_keywords()

        def fix_key(key):
            """
            Make HIERARCH keywords and fix Template parameters.
            """
            if "__" in key:
                # Workaround for template parameters
                elements = [kk for kk in key.split() if "__" in kk]
                assert len(elements) == 1
                element = elements[0]
                key = element.replace("__", " ")
            if " " in key:
                key = f'HIERARCH ESO {key}'
            return key
        kws2 = OrderedDict([
            (fix_key(k), v)
            for k, v in kws.items()
        ])
        return kws2

    def real_fits_keywords(self, ppp=None):
        """Real FITS kws."""
        from micado.main.Hardware import Hardware
        from micado.main.Template import Template
        from micado.main.ObservingBlock import ObservingBlock
        from micado.main.Simulator import SimulatorRun
        from micado.main.DataItem import DataItem
        from micado.main.Recipe import Recipe

        classes_ok = (Hardware, Template, ObservingBlock, Recipe, DataItem)

        ppp = [] if ppp is None else ppp
        proptypes1 = {
            an: getattr(self.__class__, an).prop_type
            for an in self.get_persistent()
        }
        proptypes = {
            # sometimes prop_type is (float, int)
            an: pt[0] if isinstance(pt, tuple) else pt
            for an, pt in proptypes1.items()
        }
        attrnames = [
            attrnamea
            for attrnamea, proptype in proptypes.items()
            if not attrnamea.startswith('dummy') and attrnamea not in [
                'object_id', 'extensions', 'globalname',
            ] and not issubclass(proptype, SimulatorRun)]

        def getkw(attrc, prefix=None):
            """
            Recursively get keyword attributes.
            """
            prefix = [] if prefix is None else prefix
            if isinstance(attrc, DataItem):
                if attrc.filename:
                    prefix2 = prefix + ["filename"]
                    ww = OrderedDict([(" ".join(prefix2).upper(), attrc.filename_short)])
                else:
                    ww = OrderedDict([(" ".join(prefix).upper(), "")])
            elif isinstance(attrc, classes_ok):
                ww = attrc.real_fits_keywords(ppp=prefix)
            else:
                ww = OrderedDict([(" ".join(prefix).upper(), attrc)])
            return ww

        attrs = OrderedDict()
        for attrname in attrnames:
            attr: MObject = self.__getattribute__(attrname)
            attrname_short = "{:.3}".format(attrname)
            # TODO: Make more general solution for this ugly TEM/TPL hack.
            if attrname_short == "tem":
                attrname_short = "tpl"
            if isinstance(attr, list):
                for attri, attr2 in enumerate(attr):
                    prefix2 = ppp + ["{}{}".format(attrname_short, attri+1)]
                    attrs.update(getkw(attr2, prefix=prefix2))
            elif isinstance(attr, classes_ok):
                prefix2 = ppp + [attrname_short]
                attrs.update(getkw(attr, prefix=prefix2))
            else:
                # Convert 'instrument' to 'INSTRUME' but keep 'det__ndit' in tact.
                attrname8 = attrname[:8] if "__" not in attrname else attrname
                prefix2 = ppp + [attrname8]
                attrs.update(getkw(attr, prefix=prefix2))

        return attrs

    @property
    def hierarch_shortnames(self):
        """
        2-way dictionary with short names for persistent attributes

        E.g. for Dark
        {'detector': 'DET',
         'observing_block': 'OBS',
         'simulator': 'SIM',
         'template': 'TEM',
         'DET': 'detector',
         'OBS': 'observing_block',
         'SIM': 'simulator',
         'TEM': 'template'}
        """
        # TODO: use this function somehowe, e.g. TEM should become TPL
        attrnames = self.__class__.get_persistent_properties()
        props = [getattr(self.__class__, an) for an in attrnames]
        attrshierarch = [
            (a, a[:3].upper()) for (a, p) in zip(attrnames, props)
            if isinstance(p, link_property)
        ]
        attrsvv = [a[::-1] for a in attrshierarch]
        dd = dict(attrshierarch + attrsvv)
        return dd

    def __eq__(self, other):
        """
        Test equality by comparing persistent properties.
        """
        if type(self) != type(other):
            return False

        attrnames = [
            a for a in self.get_persistent()
            if not a.startswith("dummy")
        ]
        for a in attrnames:
            v1 = getattr(self, a)
            v2 = getattr(other, a)
            if v1 != v2:
                return False

        return True
