"""
Module with persistent Raw classes.
"""

import socket
from datetime import datetime
from typing import List

from astropy.io import fits

import scopesim

from common.database.DBMain import persistent
from micado.main.Hardware import Hardware
from micado.main.BaseFrame import BaseFrame
from micado.main.ObservingBlock import ObservingBlock
from micado.main.Simulator import SimulatorRun


class Raw(BaseFrame):
    """
    A Raw DataItem.
    """

    observing_block: ObservingBlock = persistent("The Observing Block used to create this Raw", ObservingBlock, None)

    # template: Template = persistent("The Template used to create this Raw, should be in the OB", Template, None)

    # TODO: Any data can be simulated, not just Raw data.
    # TODO: reenable reading from FITS headers once they meet the standard (or vice versa)
    # simulator: SimulatorRun = persistent('The simulator producing this frame', SimulatorRun, None)

    def __init__(self, *args, **kwargs):
        """
        Set detector from extensions if necessary.
        """
        super().__init__(*args, **kwargs)

    @property
    def scopesim_dictionaries(self):
        """
        Generate ScopeSIM dictionaries.
        """

        props_hw = {
            name: self.__getattribute__(name)
            for name in self.get_persistent()
            if isinstance(self.__getattribute__(name), Hardware)
        }

        dicts = [
            hw.scopesim_dictionary
            for name, hw in props_hw.items()
        ]

        return dicts

    def make_simulation(self):
        """
        Simulate data using ScopeSIM.
        """

        # TODO: ensure observing_block actually has a role
        # obsblock_dict = self.observing_block.scopesim_dictionary
        # template_dict = obsblock_dict['templates'][0]
        template_dict0 = self.observing_block.templates[0].scopesim_dictionary

        template_dict = {
            # TODO: Ensure ScopeSIM doesn't need OBS_
            f"!OBS.{a.split('__')[-1]}": v
            for a, v in template_dict0['kwargs'].items()
            if a in ["det__dit", "det__ndit"]
        }

        # TODO: Generalize these scopesim dictionaries even more.
        observations_dict = {
            "!INST.pixel_scale": 0.004,  # because optical train still need this
            "!INST.plate_scale": 0.2666667,  # because optical train still need this
        }

        observations_dict.update(template_dict)

        import pprint
        pprint.pprint(observations_dict)
        pprint.pprint(self.scopesim_dictionaries)

        target = self.observing_block.target
        src_function = getattr(scopesim.source.source_templates, target)

        self.simulator = SimulatorRun()
        self.simulator.run_date = datetime.now()
        self.simulator.machine_dns = socket.getfqdn()

        src = src_function()
        cmd = scopesim.UserCommands(
            yamls=self.scopesim_dictionaries,
            properties=observations_dict,
        )
        opt = scopesim.OpticalTrain(cmd)
        opt.observe(src)
        hdulist = opt.readout()

        # TODO: reenable simulator, maybe pointing the other way?
        # duration = datetime.now() - self.simulator.run_date
        # self.simulator.run_duration = duration.total_seconds()

        # TODO: reenable reading from FITS headers once they meet the standard (or vice versa)
        # self.load_hdulist(hdulist)

        hdus: List[fits.ImageHDU] = [hdu for hdu in hdulist[1:] if isinstance(hdu, fits.ImageHDU)]
        for extension, hdu in zip(self.extensions, hdus):
            extension.image = hdu.data

