# -*- coding: utf-8 -*-
"""
Tests for MasterDark.py
"""

import os

from micado.config.Environment import Env
from micado.main.ObservingBlock import ObservingBlock
from micado.hardware.Detector import Detector
from micado.hardware.DetectorArray import DetectorArray
from micado.hardware.DarkCurrent import DarkCurrent
from micado.dataitems.Dark import Dark
from micado.dataitems.Dark import DarkExtension
from micado.dataitems.MasterDark import MasterDark
from micado.dataitems.MasterDark import MasterDarkExtension
from micado.main.Imstat import Imstat
from micado.requirements.DPS3000SCI import DPS3000SCI
from micado.templates.MCD_img_cal_dark import MCD_img_cal_dark
from micado.recipes.McdDetDark import McdDetDark


def create_masterdark():
    """
    Create MasterDark for testing.
    TODO: Replace with fixture?
    """
    # Set the real dark current and create a DetectorArray.
    dark_real = 100.
    mydetector = Detector()
    mydetectorlist = [mydetector]
    mydarkcurrent = DarkCurrent(value=dark_real)
    mydetector = DetectorArray(
        name='testdetector',
        detector_array_list=mydetectorlist,
        dark_current=mydarkcurrent,
    )

    # A template
    dark_template = MCD_img_cal_dark(det__dit=1, det__ndit=1)

    # An observingblock
    dark_ob = ObservingBlock(
        name="MyDarkTest",
        instrument="MICADO",
        target="empty_sky",
        templates=[dark_template],
    )

    # DarkExtensions
    myextensions = [
        DarkExtension(detector=det, BITPIX=-64, NAXIS=2)
        for det in mydetectorlist
    ]

    # Create raw Darks.
    n_darks = 5
    mydarks = []
    for i in range(n_darks):
        mydark = Dark(
            detector=mydetector,
            template=dark_template,
            observing_block=dark_ob,
            extensions=myextensions,
        )
        mydark.make_simulation()
        mydark.save_fits()
        mydarks.append(mydark)

    # A recipe. The dark recipe has no parameters.
    myrecipe = McdDetDark()

    mymasterextensions = [
        MasterDarkExtension(BITPIX=-64, NAXIS=2)
        for _ in myextensions
    ]

    # Create MasterDark.
    mymasterdark = MasterDark(
        recipe=myrecipe,
        darks=mydarks,
        extensions=mymasterextensions,
    )
    mymasterdark.make()
    mymasterdark.save_fits()
    return mydarks, mymasterdark, dark_real


def test_processing():
    """
    Test creation of MasterDark.
    """
    Env['use_python_logging'] = True
    mydarks, mymasterdark, dark_real = create_masterdark()

    # Create image statistics.
    myimstat = Imstat(baseframe=mymasterdark)
    myimstat.make()

    # Verify that the measured dark current is the same as the real current.
    myreq = DPS3000SCI()
    myreq.dark_true = dark_real
    myreq.imstat_masterdark = myimstat
    myreq.make()
    assert myreq.passed, "Requirement not met!"

    myreqfail = DPS3000SCI()
    myreqfail.dark_true = dark_real * 8
    myreqfail.imstat_masterdark = myimstat
    myreqfail.make()
    assert not myreqfail.passed, "Requirement met inadvertently."

    fn = mymasterdark.filename

    master2 = MasterDark(filename=fn)
    assert mymasterdark.fits_keywords == master2.fits_keywords
    assert mymasterdark.recipe == master2.recipe

    for d1, d2 in zip(mymasterdark.darks, master2.darks):
        assert d1.fits_keywords == d2.fits_keywords
        assert d1 == d2

    ext1, ext2 = mymasterdark.extensions[0], master2.extensions[0]
    assert ext1.fits_keywords == ext2.fits_keywords
    assert ext1 == ext2
    assert mymasterdark == master2

    for dm in mydarks:
        os.remove(dm.filename)
        os.remove(dm.filename_short)

    os.remove(mymasterdark.filename)
    os.remove(mymasterdark.filename_short)
