from dataclasses import dataclass, field
from typing import List


class Art:
    color = "trans"

    @property
    def lines(self):
        return ['']

    @property
    def height(self):
        return len(self.lines)

    @property
    def width(self):
        return len(self.lines[0])

    def __str__(self):
        ss = "".join(s + "\n" for s in self.lines)
        return ss


@dataclass
class PosArt:
    x: int = 0
    y: int = 0
    art: Art = None


@dataclass
class Canvas(Art):
    arts: List[PosArt]
    empty: str = " "
    color: str = "trans"

    @property
    def lines(self):
        lines = ['']
        for pa in self.arts:
            lines = self._add(lines, pa)
        return lines

    def _add(self, lines, pa: PosArt):
        while len(lines) < pa.y + pa.art.height:
            lines.append(self.empty * len(lines[0]))
        while len(lines[0]) < pa.x + pa.art.width:
            lines = [
                l + self.empty for l in lines
            ]
        for i, l in enumerate(pa.art.lines):
            l2 = lines[pa.y + i]
            la = "".join(
                cold if (cnew in (' ', self.empty)) & (pa.art.color == 'trans') else cnew
                for cold, cnew in zip(l2[pa.x:pa.x + pa.art.width], l)
            )
            l3 = l2[:pa.x] + la + l2[pa.x + pa.art.width:]

            lines[pa.y + i] = l3
        return lines

    def __str__(self):
        ss = "".join(l + "\n" for l in self.lines)
        return ss


@dataclass
class Box(Art):
    text: str = ""
    color: str = "   "
    rounded: bool = False
    tag: str = ""

    @property
    def lines(self):
        tcolor = "    " if self.color == "trans" else self.color
        sin = self.text + " "
        lsin = len(sin) + 1
        ltag = len(self.tag)
        ltop = "/" + "-" * lsin + "\\" if self.rounded else "+" + "-" * lsin + "+"
        l2 = "|" + ((f"{{{self.tag}}}" + " " * (lsin - ltag - 2)) if self.tag else " " * lsin) + "|"
        lmid = "| " + sin + "|"
        l4 = "|" + tcolor + " " * (lsin - len(tcolor)) + "|"
        lbot = "\\" + "-" * lsin + "/" if self.rounded else "+" + "-" * lsin + "+"
        lines = [ltop, l2, lmid, l4, lbot]
        return lines

    def __str__(self):
        ss = "".join(s + "\n" for s in self.lines)
        return ss


@dataclass
class Process(Box):
    """
    A processing step.
    """
    # red https://personal.sron.nl/~pault/
    color: str = "cB56"
    rounded: bool = True


@dataclass
class Data(Box):
    """
    Input or Output data
    """
    color: str = "cBLK"
    rounded: bool = False


@dataclass
class DataRaw(Data):
    """
    Raw data used by this pipeline.
    """
    color: str = "c999"


@dataclass
class DataStatic(Data):
    """
    Static calibration data. Produced with data from another observation.
    """
    # yellow https://personal.sron.nl/~pault/
    color: str = "cCB4"


@dataclass
class DataIntermediate(Data):
    """
    Data created and used by this Pipeline.
    """
    # blue https://personal.sron.nl/~pault/
    color: str = "c47A"


@dataclass
class DataInternal(Data):
    """
    Data used internally by this recipe. Not used atm.
    """
    # orange https://personal.sron.nl/~pault/
    color: str = "cE73"


@dataclass
class DataExternal(Box):
    """
    Calibration data produced externally. That is, persistence.
    """
    # cyan https://personal.sron.nl/~pault/
    color: str = "c6CE"


@dataclass
class DataFinal(Data):
    """
    Final data produced by our pipelines to be used by scientists.
    """
    # green https://personal.sron.nl/~pault/
    color: str = "c283"


@dataclass
class Template(Box):
    """
    A Template
    """
    color: str = "cFFF"


@dataclass
class RecipeBox(Box):
    width: int = 0
    height: int = 0
    color: str = "trans"
    rounded: bool = False

    @property
    def lines(self):
        win = self.width - 2
        hin = self.height - 4
        ltop = "/" + "-" * win + "\\" if self.rounded else "+" + "-" * win + "+"
        lname = ":" + self.text + " " * (win - len(self.text)) + "|"
        lmid = "|" + " " * win + "|"
        ltag = "|" + ((f"{{{self.tag}}}" + " " * (win - len(self.tag) - 2)) if self.tag else " " * win) + "|"
        lbot = "\\" + "-" * win + "/" if self.rounded else "+" + "-" * win + "+"
        lines = [ltop, lname] + hin * [lmid] + [ltag, lbot]
        return lines


@dataclass
class ParamsBox(Box):
    """
    Process Parameters
    """
    text: List[str] = field(default_factory=list)
    color: str = "trans"
    rounded: bool = False

    @property
    def lines(self):
        tcolor = "    " if self.color == "trans" else self.color
        width_text = max([len(t) for t in self.text])
        lsin = width_text + 2
        ltop = "/" + "-" * lsin + "\\" if self.rounded else "+" + "-" * lsin + "+"
        lstext = [
            f"| {t:{width_text}} |" for t in self.text
        ]
        lbot = "\\" + "-" * lsin + "/" if self.rounded else "+" + "-" * lsin + "+"
        lines = [ltop] + lstext + [lbot]
        return lines


@dataclass
class Arrow(Art):
    vertical: bool = False
    distance: int = 3
    color: str = "trans"

    @property
    def lines(self):
        l2 = self.distance - 2
        if not self.vertical:
            if self.distance > 0:
                lines = ['+' + '-' * l2 + ">"]
            else:
                lines = ['<' + '-' * l2 + "+"]
        else:
            if self.distance > 0:
                lines = ['+'] + ['|'] * l2 + ['v']
            else:
                lines = ['^'] + ['|'] * l2 + ['+']
        return lines


class Circle(Art):
    color = "cBLK"

    @property
    def lines(self):
        lines = [
            "+-----+",
            "|     |",
            "|{o}  |",
            "|cBLK |",
            "+-----+",
        ]
        return lines


@dataclass
class RecipeStep(Art):
    # TODO: inherit from Canvas directly
    process: Process = None
    inputs: List[Box] = None
    params: ParamsBox = None
    raw: DataRaw = None
    previous: Process = None
    xprocess: int = 10
    xarrow: int = 20
    ws: int = 30
    color: str = "trans"

    @property
    def lines(self):
        xprocessstep = self.xprocess + (self.ws - self.process.width) // 2
        posarts = []
        y = 2 if len(self.inputs) else 3
        for inp in self.inputs:
            pa = PosArt(0, y, inp)
            yo = y
            y += inp.height + 1
            arr = Arrow(vertical=False, distance=self.xarrow - inp.width + 1)
            paarr = PosArt(inp.width - 1, yo + 2, arr)
            posarts.append(pa)
            posarts.append(paarr)

        if self.params is not None:
            assert self.params.height > 2
            if self.params.height == 3:
                paparams = PosArt(0, y+1, self.params)
            elif self.params.height in [4, 5]:
                paparams = PosArt(0, y, self.params)
            else:
                paparams = PosArt(0, y, self.params)
                y += self.params.height - 5
            posarts.append(paparams)
            pararr = Arrow(
                vertical=False,
                distance=xprocessstep - self.params.width + 1
            )
            popararr = PosArt(
                self.params.width - 1,
                y + 2,
                pararr
            )
            posarts.append(popararr)

        posarts += [
            PosArt(xprocessstep, y, self.process),
            PosArt(self.xarrow, 0, Arrow(vertical=True, distance=y))
        ]
        c = Canvas(posarts)
        return c.lines

    @property
    def lines_selfcontained(self):
        """
        Get selfcontained version of recipestep.
        """
        xprocessstep = self.xprocess + (self.ws - self.process.width) // 2
        posarts = []
        if self.raw is not None:
            prev = self.raw
        elif self.previous is not None:
            prev = self.previous
        else:
            raise ValueError("Need either raw or previous.")

        posarts.append(PosArt(xprocessstep + (self.process.width - prev.width) // 2, 0, prev))
        posarts.append(PosArt(0, prev.height - 1, self))
        c = Canvas(posarts)
        return c.lines


@dataclass
class RecipeFull(Art):
    template: Template = None
    raw: DataRaw = None
    recipe: RecipeBox = None
    steps: List[RecipeStep] = None
    product: Data = None
    color: str = "trans"

    @property
    def lines(self):
        # TODO: use pydantic dataclasses and do this correction in __post_init__
        wis = [inp.width for step in self.steps for inp in step.inputs] + \
            [step.params.width for step in self.steps if step.params is not None]
        wi = max(wis)
        wss = [step.process.width for step in self.steps]
        ws = max(wss)
        if len(self.recipe.text) > ws // 2:
            # The name of the recipe is so large that it will go over the
            # down arrow going into the recipe.
            xprocess = wi + 3
            xarrow = xprocess + len(self.recipe.text)
            wsn = (xarrow - xprocess) * 2
            width_recipe = xarrow - xprocess + ws // 2 + 4
        else:
            # TODO: Create a test with Detrend recipe.
            # Move the processteps one character to the right to make room
            # for the incoming process parameters arrow.
            xprocess = wi + 4
            xarrow = xprocess + ws // 2
            wsn = ws
            width_recipe = wsn + 5
        for step in self.steps:
            step.xprocess = xprocess
            step.xarrow = xarrow
            step.ws = wsn
        self.recipe.width = width_recipe

        y = 0
        pas = list()
        pas.append(PosArt(xprocess + (wsn - self.template.width) // 2, y, self.template))
        y += self.template.height
        arr = Arrow(vertical=True, distance=9)
        pas.append(PosArt(xarrow, y - 1, arr))
        pas.append(PosArt(xarrow-3, y + 1, Circle()))
        y += arr.height - 1
        pas.append(PosArt(xprocess + (wsn - self.raw.width) // 2, y, self.raw))
        y += self.raw.height
        pas.append(PosArt(xarrow, y-1, Arrow(vertical=True, distance=3)))
        yrb = y + 1
        y = yrb + 0
        for step in self.steps:
            pas.append(PosArt(0, y, step))
            y = y + step.height - 1

        self.recipe.height = y - yrb + 3
        arr = Arrow(vertical=True, distance=7)
        pas.append(PosArt(xarrow, y, arr))
        y += arr.height
        arr = Arrow(vertical=False, distance=self.recipe.width + wi - xarrow + 2)
        pas.append(PosArt(xarrow, y-3, arr))
        pas.append(PosArt(self.recipe.width + wi + 2, y-5, self.product))
        pas.append(PosArt(xarrow-3, y, Circle()))

        pas.append(PosArt(wi + 1, yrb, self.recipe))

        ca = Canvas(arts=pas)
        return ca.lines


# TODO: use pydantic dataclasses to harmonize these with the above.
@dataclass
class SimpleStepParams:
    text: List[str] = field(default_factory=list)


@dataclass
class SimpleStepInput:
    cls: type = None
    text: str = ""


@dataclass
class SimpleStep:
    process: str = ""
    inputs: List[SimpleStepInput] = None
    params: SimpleStepParams = None
    raw: DataRaw = None
    previous: "SimpleStep" = None


@dataclass
class SimpleRecipe:
    template: str = ""
    raw: str = ""
    recipe: str = ""
    steps: List[SimpleStep] = None
    product: str = ""


def testrecipe():
    sr = SimpleRecipe(
        template="MCD_img_sci_object",
        raw="IMAGE_RAW",
        # recipe="mcd",
        recipe="mcd_img_detrend",
        steps=[
            SimpleStep(
                process="AddWeightQualityLayer",
                inputs=[SimpleStepInput(cls=DataStatic, text="ILLUM_IMG")],
                params = SimpleStepParams(
                    text=["QI"],
                )
            ), SimpleStep(
                process="CorrectAdditive",
                inputs=[
                    SimpleStepInput(cls=DataIntermediate, text="MASTERDARK"),
                    SimpleStepInput(cls=DataExternal, text="PERSISTENCE"),
                ],
                params=SimpleStepParams(
                    text=["THRESH_LOW", "THRESH_UP_VERYLONG", "A", "B", "C"],
                )
            ), SimpleStep(
                process="UpdateHeader",
                inputs=[],
                params=SimpleStepParams(text=["P", "R"],)
            ),
        ],
        product="DETREND_IMG",
    )

    asteps = [
        RecipeStep(
            process=Process(step.process),
            inputs=[n.cls(n.text) for n in step.inputs],
            params=ParamsBox(text=step.params.text) if step.params is not None else None
        ) for step in sr.steps
    ]

    rf = RecipeFull(
        template=Template(sr.template),
        raw=DataRaw(sr.raw),
        steps=asteps,
        recipe=RecipeBox(text=sr.recipe),
        product=DataFinal(sr.product)
    )
    return rf


if __name__ == '__main__':
    tr = testrecipe()
    print(tr)
